from emmet.api.routes.electrodes.utils import (
    electrodes_chemsys_to_criteria,
    electrodes_formula_to_criteria,
)
import pytest


def test_electrodes_formula_to_criteria():
    # Regular formula
    assert electrodes_formula_to_criteria("Cr2O3") == {
        "entries_composition_summary.all_composition_reduced.Cr": 2.0,
        "entries_composition_summary.all_composition_reduced.O": 3.0,
        "nelements": {"$in": [2, 1]},
    }

    assert electrodes_formula_to_criteria("Cr2O3, FeLiO4P") == {
        "entries_composition_summary.all_formulas": {"$in": ["Cr2O3", "LiFePO4"]}
    }
    # Add wildcard
    assert electrodes_formula_to_criteria("Cr2*3") == {
        "entries_composition_summary.all_composition_reduced.Cr": 2.0,
        "entries_composition_summary.all_formula_anonymous": "A2B3",
    }
    # Anonymous element
    assert electrodes_formula_to_criteria("A2B3") == {
        "entries_composition_summary.all_formula_anonymous": "A2B3"
    }

    assert electrodes_formula_to_criteria("A2B3, ABC3") == {
        "entries_composition_summary.all_formula_anonymous": {"$in": ["A2B3", "ABC3"]}
    }


def test_electrodes_chemsys_to_criteria():
    # Chemsys
    assert electrodes_chemsys_to_criteria("Si-O") == {
        "entries_composition_summary.all_chemsys": "O-Si"
    }
    assert electrodes_chemsys_to_criteria("Si-*") == {
        "entries_composition_summary.all_elements": {"$all": ["Si"]},
        "nelements": {"$in": [2, 1]},
    }
    assert electrodes_chemsys_to_criteria("*-*-*") == {"nelements": {"$in": [3, 2]}}

    assert electrodes_chemsys_to_criteria("Si-O, P-Li-Fe") == {
        "entries_composition_summary.all_chemsys": {"$in": ["O-Si", "Fe-Li-P"]}
    }


@pytest.mark.xfail()
def test_electrodes_chemsys_to_criteria_multiple_wildcard():
    electrodes_chemsys_to_criteria("Si-O, Li-Fe-*")


@pytest.mark.xfail()
def test_electrodes_formula_to_criteria_multiple_wildcard():
    electrodes_formula_to_criteria("SiO, LiFe*")



import os
from emmet.api.routes.tasks.query_operators import (
    MultipleTaskIDsQuery,
    TrajectoryQuery,
    DeprecationQuery,
    EntryQuery,
)
from emmet.api.core.settings import MAPISettings

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn
from json import load


def test_multiple_task_ids_query():
    op = MultipleTaskIDsQuery()

    assert op.query(task_ids=" mp-149, mp-13") == {
        "criteria": {"task_id": {"$in": ["mp-149", "mp-13"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")

        assert new_op.query(task_ids=" mp-149, mp-13") == {
            "criteria": {"task_id": {"$in": ["mp-149", "mp-13"]}}
        }


def test_entries_query():
    op = EntryQuery()

    assert op.query(task_ids=" mp-149, mp-13") == {
        "criteria": {"task_id": {"$in": ["mp-149", "mp-13"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        query = {"criteria": {"task_id": {"$in": ["mp-149", "mp-13"]}}}

        assert new_op.query(task_ids=" mp-149, mp-13") == query

    with open(os.path.join(MAPISettings().TEST_FILES, "tasks_Li_Fe_V.json")) as file:
        tasks = load(file)
    docs = op.post_process(tasks, query)
    assert docs[0]["entry"]["@class"] == "ComputedStructureEntry"


def test_trajectory_query():
    op = TrajectoryQuery()

    assert op.query(task_ids=" mp-149, mp-13") == {
        "criteria": {"task_id": {"$in": ["mp-149", "mp-13"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        query = {"criteria": {"task_id": {"$in": ["mp-149", "mp-13"]}}}

        assert new_op.query(task_ids=" mp-149, mp-13") == query

    with open(os.path.join(MAPISettings().TEST_FILES, "tasks_Li_Fe_V.json")) as file:
        tasks = load(file)
    docs = op.post_process(tasks, query)
    assert docs[0]["trajectories"][0]["@class"] == "Trajectory"


def test_deprecation_query():
    op = DeprecationQuery()

    assert op.query(task_ids=" mp-149, mp-13") == {
        "criteria": {"deprecated_tasks": {"$in": ["mp-149", "mp-13"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        query = {"criteria": {"deprecated_tasks": {"$in": ["mp-149", "mp-13"]}}}

        assert new_op.query(task_ids=" mp-149, mp-13") == query

    docs = [
        {"task_id": "mp-149", "deprecated_tasks": ["mp-149"]},
        {"task_id": "mp-13", "deprecated_tasks": ["mp-1234"]},
    ]
    r = op.post_process(docs, query)

    assert r[0] == {
        "task_id": "mp-149",
        "deprecated": True,
        "deprecation_reason": None,
    }

    assert r[1] == {
        "task_id": "mp-13",
        "deprecated": False,
        "deprecation_reason": None,
    }



from emmet.api.routes.electrodes.query_operators import (
    ElectrodeFormulaQuery,
    ElectrodesChemsysQuery,
    ElectrodeElementsQuery,
    VoltageStepQuery,
    InsertionVoltageStepQuery,
    WorkingIonQuery,
    ElectrodeMultiMaterialIDQuery,
    MultiBatteryIDQuery,
)

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_electrode_formula_query():
    op = ElectrodeFormulaQuery()

    assert op.query(formula="BiFeO3") == {
        "criteria": {
            "entries_composition_summary.all_composition_reduced.Bi": 1.0,
            "entries_composition_summary.all_composition_reduced.Fe": 1.0,
            "entries_composition_summary.all_composition_reduced.O": 3.0,
            "nelements": {"$in": [3, 2]},
        }
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(formula="BiFeO3") == {
            "criteria": {
                "entries_composition_summary.all_composition_reduced.Bi": 1.0,
                "entries_composition_summary.all_composition_reduced.Fe": 1.0,
                "entries_composition_summary.all_composition_reduced.O": 3.0,
                "nelements": {"$in": [3, 2]},
            }
        }


def test_electrodes_chemsys_query():
    op = ElectrodesChemsysQuery()
    assert op.query("Si-O") == {"criteria": {"entries_composition_summary.all_chemsys": "O-Si"}}

    assert op.query("Si-*") == {
        "criteria": {"nelements": {"$in": [2, 1]}, "entries_composition_summary.all_elements": {"$all": ["Si"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query("Si-O") == {"criteria": {"entries_composition_summary.all_chemsys": "O-Si"}}


def test_electrodes_elements_query():
    eles = ["Si", "O"]
    neles = ["N", "P"]

    op = ElectrodeElementsQuery()
    assert op.query(elements=",".join(eles), exclude_elements=",".join(neles)) == {
        "criteria": {"entries_composition_summary.all_elements": {"$all": ["Si", "O"], "$nin": ["N", "P"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(elements=",".join(eles), exclude_elements=",".join(neles)) == {
            "criteria": {"entries_composition_summary.all_elements": {"$all": ["Si", "O"], "$nin": ["N", "P"]}}
        }


def test_voltage_step_query():
    op = VoltageStepQuery()

    q = op.query(
        delta_volume_min=0,
        delta_volume_max=5,
        average_voltage_min=0,
        average_voltage_max=5,
        max_voltage_min=0,
        max_voltage_max=5,
        min_voltage_min=0,
        min_voltage_max=5,
        capacity_grav_min=0,
        capacity_grav_max=5,
        capacity_vol_min=0,
        capacity_vol_max=5,
        energy_grav_min=0,
        energy_grav_max=5,
        energy_vol_min=0,
        energy_vol_max=5,
        fracA_charge_min=0,
        fracA_charge_max=5,
        fracA_discharge_min=0,
        fracA_discharge_max=5,
    )

    fields = [
        "max_delta_volume",
        "average_voltage",
        "max_voltage",
        "min_voltage",
        "capacity_grav",
        "capacity_vol",
        "energy_grav",
        "energy_vol",
        "fracA_charge",
        "fracA_discharge",
    ]

    assert q == {"criteria": {field: {"$gte": 0, "$lte": 5} for field in fields}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        q = new_op.query(
            delta_volume_min=0,
            delta_volume_max=5,
            average_voltage_min=0,
            average_voltage_max=5,
            max_voltage_min=0,
            max_voltage_max=5,
            min_voltage_min=0,
            min_voltage_max=5,
            capacity_grav_min=0,
            capacity_grav_max=5,
            capacity_vol_min=0,
            capacity_vol_max=5,
            energy_grav_min=0,
            energy_grav_max=5,
            energy_vol_min=0,
            energy_vol_max=5,
            fracA_charge_min=0,
            fracA_charge_max=5,
            fracA_discharge_min=0,
            fracA_discharge_max=5,
        )
        assert q == {"criteria": {field: {"$gte": 0, "$lte": 5} for field in fields}}


def test_insertion_voltage_step_query():
    op = InsertionVoltageStepQuery()

    q = op.query(stability_charge_min=0, stability_charge_max=5, stability_discharge_min=0, stability_discharge_max=5,)

    fields = [
        "stability_charge",
        "stability_discharge",
    ]

    assert q == {"criteria": {field: {"$gte": 0, "$lte": 5} for field in fields}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        q = new_op.query(
            stability_charge_min=0, stability_charge_max=5, stability_discharge_min=0, stability_discharge_max=5,
        )
        assert q == {"criteria": {field: {"$gte": 0, "$lte": 5} for field in fields}}


def test_insertion_electrode_query():
    op = WorkingIonQuery()

    q = op.query(working_ion="Li",)

    assert q == {"criteria": {"working_ion": "Li"}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        q = new_op.query(working_ion="Li",)

        assert q == {"criteria": {"working_ion": "Li"}}


def test_electrodes_multi_material_id_query():
    op = ElectrodeMultiMaterialIDQuery()
    assert op.query(material_ids="mp-149, mp-13") == {"criteria": {"material_ids": {"$in": ["mp-149", "mp-13"]}}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(material_ids="mp-149, mp-13") == {
            "criteria": {"material_ids": {"$in": ["mp-149", "mp-13"]}}
        }


def test_multi_battery_id_query():
    op = MultiBatteryIDQuery()
    assert op.query(battery_ids="mp-149_Ca, mp-13_Li") == {
        "criteria": {"battery_id": {"$in": ["mp-149_Ca", "mp-13_Li"]}}
    }

    assert op.query(battery_ids="mp-149_Ca") == {"criteria": {"battery_id": "mp-149_Ca"}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(battery_ids="mp-149_Ca, mp-13_Li") == {
            "criteria": {"battery_id": {"$in": ["mp-149_Ca", "mp-13_Li"]}}
        }

        assert new_op.query(battery_ids="mp-149_Ca") == {"criteria": {"battery_id": "mp-149_Ca"}}



from emmet.api.routes.materials.hint_scheme import MaterialsHintScheme


def test_materials_hint_scheme():
    scheme = MaterialsHintScheme()
    assert scheme.generate_hints({"criteria": {"nelements": 3}}) == {
        "hint": {"nelements": 1}
    }



import os
from emmet.api.routes.mpcomplete.query_operator import (
    MPCompleteGetQuery,
    MPCompletePostQuery,
)
from emmet.api.core.settings import MAPISettings

from pymatgen.core.structure import Structure

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_mpcomplete_post_query():
    op = MPCompletePostQuery()

    structure = Structure.from_file(
        os.path.join(MAPISettings().TEST_FILES, "Si_mp_149.cif")
    )

    assert op.query(
        structure=structure.as_dict(),
        public_name="Test Test",
        public_email="test@test.com",
    ) == {
        "criteria": {
            "structure": structure.as_dict(),
            "public_name": "Test Test",
            "public_email": "test@test.com",
        }
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        query = {
            "criteria": {
                "structure": structure.as_dict(),
                "public_name": "Test Test",
                "public_email": "test@test.com",
            }
        }
        assert new_op.query(
            structure=structure.as_dict(),
            public_name="Test Test",
            public_email="test@test.com",
        ) == query

    docs = [
        {
            "structure": structure.as_dict(),
            "public_name": "Test Test",
            "public_email": "test@test.com",
        }
    ]
    assert op.post_process(docs, query) == docs


def test_mocomplete_get_query():
    op = MPCompleteGetQuery()

    assert (
        op.query(
            public_name="Test Test",
            public_email="test@test.com",
        )
        == {"criteria": {"public_name": "Test Test", "public_email": "test@test.com"}}
    )

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert (
            new_op.query(
                public_name="Test Test",
                public_email="test@test.com",
            )
            == {
                "criteria": {
                    "public_name": "Test Test",
                    "public_email": "test@test.com",
                }
            }
        )



from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn

from emmet.api.routes.summary.query_operators import (
    HasPropsQuery,
    MaterialIDsSearchQuery,
    SearchHasReconstructedQuery,
    SearchIsStableQuery,
    SearchMagneticQuery,
    SearchIsTheoreticalQuery,
    SearchStatsQuery,
    SearchESQuery,
)

from emmet.core.summary import SummaryStats

from pymatgen.analysis.magnetism import Ordering
from emmet.core.summary import SummaryDoc


def test_has_props_query():
    op = HasPropsQuery()

    assert op.query(has_props="electronic_structure, thermo") == {
        "criteria": {"has_props": {"$all": ["electronic_structure", "thermo"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(has_props="electronic_structure, thermo") == {
            "criteria": {"has_props": {"$all": ["electronic_structure", "thermo"]}}
        }


def test_material_ids_query():
    op = MaterialIDsSearchQuery()

    query = {"criteria": {"material_id": {"$in": ["mp-149", "mp-13"]}}}

    assert op.query(material_ids="mp-149, mp-13") == query

    docs = [{"material_id": "mp-13"}, {"material_id": "mp-149"}]

    assert op.post_process(docs, {**query, "properties": ["material_id"]})[0] == docs[1]

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(material_ids="mp-149, mp-13") == query


def test_is_stable_query():
    op = SearchIsStableQuery()

    assert op.query(is_stable=True) == {"criteria": {"is_stable": True}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(is_stable=True) == {"criteria": {"is_stable": True}}


def test_magnetic_query():
    op = SearchMagneticQuery()

    assert op.query(ordering=Ordering.FiM) == {"criteria": {"ordering": "FiM"}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(ordering=Ordering.FiM) == {"criteria": {"ordering": "FiM"}}


def test_has_reconstructed_query():
    op = SearchHasReconstructedQuery()

    assert op.query(has_reconstructed=False) == {"criteria": {"has_reconstructed": False}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(has_reconstructed=False) == {"criteria": {"has_reconstructed": False}}


def test_is_theoretical_query():
    op = SearchIsTheoreticalQuery()

    assert op.query(theoretical=False) == {"criteria": {"theoretical": False}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(theoretical=False) == {"criteria": {"theoretical": False}}


def test_search_stats_query():
    op = SearchStatsQuery(SummaryDoc)

    pipeline = [
        {"$match": {"band_gap": {"$gte": 0, "$lte": 5}}},
        {"$sample": {"size": 10}},
        {"$project": {"band_gap": 1, "_id": 0}},
    ]

    assert op.query(field="band_gap", num_samples=10, min_val=0, max_val=5, num_points=100) == {"pipeline": pipeline}

    docs = [{"band_gap": 1}, {"band_gap": 2}, {"band_gap": 3}]

    assert isinstance(op.post_process(docs, {"pipeline": pipeline})[0], SummaryStats)


def test_search_es_query():
    op = SearchESQuery()

    assert op.query(is_gap_direct=False, is_metal=False) == {"criteria": {"is_gap_direct": False, "is_metal": False}}



from emmet.api.routes._consumer.query_operator import (
    UserSettingsGetQuery,
    UserSettingsPostQuery,
)

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_user_settings_post_query():
    op = UserSettingsPostQuery()

    assert op.query(consumer_id="test", settings={"test": "test", "test2": 10}) == {
        "criteria": {"consumer_id": "test", "settings": {"test": "test", "test2": 10}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        query = {
            "criteria": {"consumer_id": "test", "settings": {"test": "test", "test2": 10},}
        }
        assert new_op.query(consumer_id="test", settings={"test": "test", "test2": 10}) == {
            "criteria": {"consumer_id": "test", "settings": {"test": "test", "test2": 10},}
        }

    docs = [{"consumer_id": "test", "settings": {"test": "test", "test2": 10}}]
    assert op.post_process(docs, query) == docs


def test_user_settings_get_query():
    op = UserSettingsGetQuery()

    assert op.query(consumer_id="test") == {"criteria": {"consumer_id": "test"}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(consumer_id="test") == {"criteria": {"consumer_id": "test"}}



from emmet.api.routes.magnetism.query_operators import MagneticQuery

from pymatgen.analysis.magnetism import Ordering

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_magnetic_query():
    op = MagneticQuery()

    q = op.query(
        ordering=Ordering.FM,
        total_magnetization_min=0,
        total_magnetization_max=5,
        total_magnetization_normalized_vol_min=0,
        total_magnetization_normalized_vol_max=5,
        total_magnetization_normalized_formula_units_min=0,
        total_magnetization_normalized_formula_units_max=5,
        num_magnetic_sites_min=0,
        num_magnetic_sites_max=5,
        num_unique_magnetic_sites_min=0,
        num_unique_magnetic_sites_max=5,
    )

    fields = [
        "total_magnetization",
        "total_magnetization_normalized_vol",
        "total_magnetization_normalized_formula_units",
        "num_magnetic_sites",
        "num_unique_magnetic_sites",
    ]

    c = {field: {"$gte": 0, "$lte": 5} for field in fields}

    assert q == {"criteria": {"ordering": "FM", **c}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        q = new_op.query(
            ordering=Ordering.FM,
            total_magnetization_min=0,
            total_magnetization_max=5,
            total_magnetization_normalized_vol_min=0,
            total_magnetization_normalized_vol_max=5,
            total_magnetization_normalized_formula_units_min=0,
            total_magnetization_normalized_formula_units_max=5,
            num_magnetic_sites_min=0,
            num_magnetic_sites_max=5,
            num_unique_magnetic_sites_min=0,
            num_unique_magnetic_sites_max=5,
        )
        c = {field: {"$gte": 0, "$lte": 5} for field in fields}

        assert q == {"criteria": {"ordering": "FM", **c}}



from emmet.api.routes._general_store.query_operator import (
    GeneralStoreGetQuery,
    GeneralStorePostQuery,
)

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_user_settings_post_query():
    op = GeneralStorePostQuery()

    assert op.query(
        kind="test", meta={"test": "test", "test2": 10}, markdown="test"
    ) == {
        "criteria": {
            "kind": "test",
            "meta": {"test": "test", "test2": 10},
            "markdown": "test",
        }
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        query = {
            "criteria": {
                "kind": "test",
                "meta": {"test": "test", "test2": 10},
                "markdown": "test",
            }
        }
        assert new_op.query(
            kind="test", meta={"test": "test", "test2": 10}, markdown="test"
        ) == query

    docs = [{"kind": "test", "meta": {"test": "test", "test2": 10}, "markdown": "test"}]
    assert op.post_process(docs, query) == docs


def test_user_settings_get_query():
    op = GeneralStoreGetQuery()

    assert op.query(kind="test") == {"criteria": {"kind": "test"}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(kind="test") == {"criteria": {"kind": "test"}}



from emmet.api.routes.materials.utils import formula_to_criteria, chemsys_to_criteria
import pytest


def test_formula_to_criteria():
    # Regular formula
    assert formula_to_criteria("Cr2O3") == {
        "composition_reduced.Cr": 2.0,
        "composition_reduced.O": 3.0,
        "nelements": 2,
    }
    assert formula_to_criteria("Cr2O3, O2Si") == {
        "formula_pretty": {"$in": ["Cr2O3", "SiO2"]}
    }

    # Add wildcard
    assert formula_to_criteria("Cr2*3") == {
        "composition_reduced.Cr": 2.0,
        "formula_anonymous": "A2B3",
    }
    # Anonymous element
    assert formula_to_criteria("A2B3") == {"formula_anonymous": "A2B3"}
    assert formula_to_criteria("A2B3, ABC3") == {
        "formula_anonymous": {"$in": ["A2B3", "ABC3"]}
    }


def test_chemsys_to_criteria():
    # Chemsys
    assert chemsys_to_criteria("Si-O") == {"chemsys": "O-Si"}
    assert chemsys_to_criteria("Si-*") == {"elements": {"$all": ["Si"]}, "nelements": 2}
    assert chemsys_to_criteria("*-*-*") == {"nelements": 3}

    assert chemsys_to_criteria("Si-O, P-Li-Fe") == {
        "chemsys": {"$in": ["O-Si", "Fe-Li-P"]}
    }


@pytest.mark.xfail()
def test_chemsys_to_criteria_multiple_wildcard():
    chemsys_to_criteria("Si-O, Li-Fe-*")


@pytest.mark.xfail()
def test_formula_to_criteria_multiple_wildcard():
    formula_to_criteria("SiO, LiFe*")



from typing import Optional
from fastapi import Query
from maggma.api.utils import STORE_PARAMS
from maggma.api.query_operator import QueryOperator


class IsStableQuery(QueryOperator):
    """
    Method to generate a query on whether a material is stable
    """

    def query(
        self,
        is_stable: Optional[bool] = Query(
            None, description="Whether the material is stable."
        ),
    ):

        crit = {}

        if is_stable is not None:
            crit["is_stable"] = is_stable

        return {"criteria": crit}

    def ensure_indexes(self):  # pragma: no cover
        keys = self._keys_from_query()
        return [(key, False) for key in keys]


class MultiThermoIDQuery(QueryOperator):
    """
    Method to generate a query for different root-level thermo_id values
    """

    def query(
        self,
        thermo_ids: Optional[str] = Query(
            None, description="Comma-separated list of thermo_id values to query on"
        ),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if thermo_ids:

            thermo_id_list = [thermo_id.strip() for thermo_id in thermo_ids.split(",")]

            if len(thermo_id_list) == 1:
                crit.update({"thermo_id": thermo_id_list[0]})
            else:
                crit.update({"thermo_id": {"$in": thermo_id_list}})

        return {"criteria": crit}


class MultiThermoTypeQuery(QueryOperator):
    """
    Method to generate a query for different root-level thermo_type values
    """

    def query(
        self,
        thermo_types: Optional[str] = Query(
            None, description="Comma-separated list of thermo_type values to query on"
        ),
    ) -> STORE_PARAMS:

        crit = {}  # type: dict

        if thermo_types:

            thermo_type_list = [
                thermo_type.strip() for thermo_type in thermo_types.split(",")
            ]

            if len(thermo_type_list) == 1:
                crit.update({"thermo_type": thermo_type_list[0]})
            else:
                crit.update({"thermo_type": {"$in": thermo_type_list}})

        return {"criteria": crit}



import os
from json import load

from emmet.api.core.settings import MAPISettings
from emmet.api.routes.tasks.utils import calcs_reversed_to_trajectory, task_to_entry


def test_calcs_reversed_to_trajectory():
    with open(
        os.path.join(MAPISettings().TEST_FILES, "calcs_reversed_mp_1031016.json")
    ) as file:
        calcs_reversed = load(file)
        trajectories = calcs_reversed_to_trajectory(calcs_reversed)

    assert len(trajectories) == 1
    assert trajectories[0]["lattice"] == [
        [9.054455, 0.0, 0.0],
        [0.0, 4.500098, 0.0],
        [0.0, 0.0, 4.500098],
    ]


def test_task_to_entry():
    with open(os.path.join(MAPISettings().TEST_FILES, "test_task.json")) as file:
        doc = load(file)

    entry_dict = task_to_entry(doc)
    assert entry_dict["@class"] == "ComputedStructureEntry"




from emmet.api.routes.xas.query_operators import XASQuery, XASTaskIDQuery

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn

from emmet.core.xas import Edge, Type
from pymatgen.core.periodic_table import Element


def test_xas_operator():
    op = XASQuery()

    assert op.query(
        edge=Edge.K, spectrum_type=Type.XANES, absorbing_element=Element("Cu")
    ) == {
        "criteria": {"edge": "K", "absorbing_element": "Cu", "spectrum_type": "XANES"}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")

        assert new_op.query(
            edge=Edge.K, spectrum_type=Type.XANES, absorbing_element=Element("Cu")
        ) == {
            "criteria": {
                "edge": "K",
                "absorbing_element": "Cu",
                "spectrum_type": "XANES",
            }
        }


def test_xas_task_id_operator():
    op = XASTaskIDQuery()

    assert op.query(material_ids="mp-149, mp-13") == {
        "criteria": {"material_id": {"$in": ["mp-149", "mp-13"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")

        assert new_op.query(material_ids="mp-149, mp-13") == {
            "criteria": {"material_id": {"$in": ["mp-149", "mp-13"]}}
        }



from typing import Optional
from fastapi import Query
from maggma.api.query_operator import QueryOperator
from maggma.api.utils import STORE_PARAMS


class MaterialIDsSearchQuery(QueryOperator):
    """
    Method to generate a query on search docs using multiple material_id values
    """

    def query(
        self,
        material_ids: Optional[str] = Query(None, description="Comma-separated list of material_ids to query on"),
    ) -> STORE_PARAMS:

        crit = {}

        if material_ids:

            terminal_search = {"_search.id": {"$in": [material_id.strip() for material_id in material_ids.split(",")]}}

            member_search = {
                "_search.member_ids": {"$in": [material_id.strip() for material_id in material_ids.split(",")]}
            }

            crit.update({"$or": [terminal_search, member_search]})

        return {"criteria": crit}


class FormulaSearchQuery(QueryOperator):
    def query(
        self,
        formulae: Optional[str] = Query(None, description="Comma-separated list of end-point formulas to query."),
    ) -> STORE_PARAMS:

        crit = {}

        if formulae:

            formula_search = {"_search.formula": {"$in": [formula.strip() for formula in formulae.split(",")]}}

            crit.update(formula_search)

        return {"criteria": crit}



from maggma.api.resource import HintScheme


class TasksHintScheme(HintScheme):
    """
    Hint scheme for the tasks endpoint.
    """

    def generate_hints(self, query):

        if query["criteria"] == {}:
            return {"hint": {"_id": 1}}

        for param in query["criteria"]:

            if "composition_reduced" in param:
                return {"hint": {"composition_reduced.$**": 1}}
            elif "nelements" in param:
                return {"hint": {"nelements": 1}}

        return {"hint": {}}



from emmet.api.routes.piezo.query_operators import PiezoelectricQuery

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_piezo_query():
    op = PiezoelectricQuery()

    assert op.query(piezo_modulus_min=0, piezo_modulus_max=5) == {
        "criteria": {"e_ij_max": {"$gte": 0, "$lte": 5}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(piezo_modulus_min=0, piezo_modulus_max=5) == {
            "criteria": {"e_ij_max": {"$gte": 0, "$lte": 5}}
        }



from emmet.api.routes.summary.hint_scheme import SummaryHintScheme


def test_summary_hint_scheme():
    scheme = SummaryHintScheme()
    assert scheme.generate_hints({"criteria": {"nelements": 3}}) == {
        "hint": {"nelements": 1}
    }
    assert scheme.generate_hints({"criteria": {"has_props": "dos"}}) == {
        "hint": {"has_props": 1}
    }



from emmet.api.routes.synthesis.query_operators import SynthesisSearchQuery
from emmet.core.synthesis import SynthesisTypeEnum, OperationTypeEnum

from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_synthesis_search_query():
    op = SynthesisSearchQuery()
    keyword_lists = [None, "silicon, process"]

    for keywords in keyword_lists:
        pipeline = (
            [
                {
                    "$match": {
                        "synthesis_type": {"$in": ["solid-state"]},
                        "targets_formula_s": "SiO2",
                        "precursors_formula_s": "SiO2",
                        "operations.type": {"$all": ["ShapingOperation"]},
                        "operations.conditions.heating_temperature.values": {
                            "$elemMatch": {"$gte": 0, "$lte": 5}
                        },
                        "operations.conditions.heating_time.values": {
                            "$elemMatch": {"$gte": 0, "$lte": 5}
                        },
                        "operations.conditions.heating_atmosphere": {"$all": ["air"]},
                        "operations.conditions.mixing_device": {"$all": ["zirconia"]},
                        "operations.conditions.mixing_media": {"$all": ["water"]},
                    }
                },
                {
                    "$project": {
                        "_id": 0,
                        "doi": 1,
                        "synthesis_type": 1,
                        "reaction": 1,
                        "reaction_string": 1,
                        "operations": 1,
                        "target": 1,
                        "targets_formula": 1,
                        "targets_formula_s": 1,
                        "precursors": 1,
                        "precursors_formula_s": 1,
                        "paragraph_string": 1,
                    }
                },
                {
                    "$facet": {
                        "results": [{"$skip": 0}, {"$limit": 10}],
                        "total_doc": [{"$count": "count"}],
                    }
                },
                {"$unwind": "$results"},
                {"$unwind": "$total_doc"},
                {
                    "$replaceRoot": {
                        "newRoot": {
                            "$mergeObjects": [
                                "$results",
                                {"total_doc": "$total_doc.count"},
                            ]
                        }
                    }
                },
            ]
            if keywords is None
            else [
                {
                    "$search": {
                        "index": "synth_descriptions",
                        "search": {
                            "query": ["silicon", "process"],
                            "path": "paragraph_string",
                        },
                        "highlight": {"path": "paragraph_string"},
                    }
                },
                {
                    "$project": {
                        "_id": 0,
                        "doi": 1,
                        "highlights": {"$meta": "searchHighlights"},
                        "synthesis_type": 1,
                        "reaction": 1,
                        "reaction_string": 1,
                        "search_score": {"$meta": "searchScore"},
                        "operations": 1,
                        "target": 1,
                        "targets_formula": 1,
                        "targets_formula_s": 1,
                        "precursors": 1,
                        "precursors_formula_s": 1,
                        "paragraph_string": 1,
                    }
                },
                {
                    "$match": {
                        "synthesis_type": {"$in": ["solid-state"]},
                        "targets_formula_s": "SiO2",
                        "precursors_formula_s": "SiO2",
                        "operations.type": {"$all": ["ShapingOperation"]},
                        "operations.conditions.heating_temperature.values": {
                            "$elemMatch": {"$gte": 0, "$lte": 5}
                        },
                        "operations.conditions.heating_time.values": {
                            "$elemMatch": {"$gte": 0, "$lte": 5}
                        },
                        "operations.conditions.heating_atmosphere": {"$all": ["air"]},
                        "operations.conditions.mixing_device": {"$all": ["zirconia"]},
                        "operations.conditions.mixing_media": {"$all": ["water"]},
                    }
                },
                {
                    "$facet": {
                        "results": [
                            {"$sort": {"search_score": -1}},
                            {"$skip": 0},
                            {"$limit": 10},
                        ],
                        "total_doc": [{"$count": "count"}],
                    }
                },
                {"$unwind": "$results"},
                {"$unwind": "$total_doc"},
                {
                    "$replaceRoot": {
                        "newRoot": {
                            "$mergeObjects": [
                                "$results",
                                {"total_doc": "$total_doc.count"},
                            ]
                        }
                    }
                },
            ]
        )

        q = op.query(
            keywords=keywords,
            synthesis_type=[SynthesisTypeEnum.solid_state],
            target_formula="SiO2",
            precursor_formula="SiO2",
            operations=[OperationTypeEnum.shaping],
            condition_heating_time_min=0,
            condition_heating_time_max=5,
            condition_heating_temperature_min=0,
            condition_heating_temperature_max=5,
            condition_heating_atmosphere=["air"],
            condition_mixing_device=["zirconia"],
            condition_mixing_media=["water"],
            _skip=0,
            _limit=10,
        )

        assert q["pipeline"] == pipeline

        with ScratchDir("."):
            dumpfn(op, "temp.json")
            new_op = loadfn("temp.json")
            q = new_op.query(
                keywords=keywords,
                synthesis_type=[SynthesisTypeEnum.solid_state],
                target_formula="SiO2",
                precursor_formula="SiO2",
                operations=[OperationTypeEnum.shaping],
                condition_heating_time_min=0,
                condition_heating_time_max=5,
                condition_heating_temperature_min=0,
                condition_heating_temperature_max=5,
                condition_heating_atmosphere=["air"],
                condition_mixing_device=["zirconia"],
                condition_mixing_media=["water"],
                _skip=0,
                _limit=10,
            )
            assert q == {"pipeline": pipeline}



from emmet.api.routes.surface_properties.query_operators import ReconstructedQuery


from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn


def test_reconstructed_operator():
    op = ReconstructedQuery()

    assert op.query(has_reconstructed=True) == {"criteria": {"has_reconstructed": True}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")

        assert new_op.query(has_reconstructed=True) == {
            "criteria": {"has_reconstructed": True}
        }



from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn

from emmet.api.routes.grain_boundary.query_operators import (
    GBStructureQuery,
    GBTaskIDQuery,
)
from emmet.core.grain_boundary import GBTypeEnum


def test_grain_boundary_structure_query():
    op = GBStructureQuery()

    assert op.query(
        sigma=5,
        type=GBTypeEnum.twist,
        chemsys="Si-Fe",
        pretty_formula="Fe2Si4",
        gb_plane="1,1,1",
        rotation_axis="1,0,1",
    ) == {
        "criteria": {
            "sigma": 5,
            "type": "twist",
            "chemsys": "Fe-Si",
            "pretty_formula": "FeSi2",
            "gb_plane": [1, 1, 1],
            "rotation_axis": [1, 0, 1],
        }
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(
            sigma=5,
            type=GBTypeEnum.twist,
            chemsys="Si-Fe",
            pretty_formula="Fe2Si4",
            gb_plane="1,1,1",
            rotation_axis="1,0,1",
        ) == {
            "criteria": {
                "sigma": 5,
                "type": "twist",
                "chemsys": "Fe-Si",
                "pretty_formula": "FeSi2",
                "gb_plane": [1, 1, 1],
                "rotation_axis": [1, 0, 1],
            }
        }


def test_grain_boundary_task_id_query():
    op = GBTaskIDQuery()

    assert op.query(task_ids="mp-149, mp-13") == {
        "criteria": {"task_id": {"$in": ["mp-149", "mp-13"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(task_ids="mp-149, mp-13") == {
            "criteria": {"task_id": {"$in": ["mp-149", "mp-13"]}}
        }



from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn
from pymatgen.electronic_structure.core import OrbitalType

from emmet.api.routes.electronic_structure.query_operators import (
    ESSummaryDataQuery,
    BSDataQuery,
    DOSDataQuery,
    ObjectQuery,
)
from emmet.core.electronic_structure import (
    BSPathType,
    DOSProjectionType,
)

from pymatgen.analysis.magnetism.analyzer import Ordering
from pymatgen.core.periodic_table import Element


def test_es_summary_query():
    op = ESSummaryDataQuery()

    assert op.query(
        magnetic_ordering=Ordering.FiM, is_gap_direct=True, is_metal=False
    ) == {
        "criteria": {
            "magnetic_ordering": "FiM",
            "is_gap_direct": True,
            "is_metal": False,
        }
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(
            magnetic_ordering=Ordering.FiM, is_gap_direct=True, is_metal=False
        ) == {
            "criteria": {
                "magnetic_ordering": "FiM",
                "is_gap_direct": True,
                "is_metal": False,
            }
        }


def test_bs_data_query():
    op = BSDataQuery()

    q = op.query(
        path_type=BSPathType.setyawan_curtarolo,
        band_gap_min=0,
        band_gap_max=5,
        efermi_min=0,
        efermi_max=5,
        magnetic_ordering=Ordering.FM,
        is_gap_direct=True,
        is_metal=False,
    )

    fields = [
        "bandstructure.setyawan_curtarolo.band_gap",
        "bandstructure.setyawan_curtarolo.efermi",
    ]

    c = {field: {"$gte": 0, "$lte": 5} for field in fields}

    assert q == {
        "criteria": {
            "bandstructure.setyawan_curtarolo.magnetic_ordering": "FM",
            "bandstructure.setyawan_curtarolo.is_gap_direct": True,
            "bandstructure.setyawan_curtarolo.is_metal": False,
            **c,
        }
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        q = new_op.query(
            path_type=BSPathType.setyawan_curtarolo,
            band_gap_min=0,
            band_gap_max=5,
            efermi_min=0,
            efermi_max=5,
            magnetic_ordering=Ordering.FM,
            is_gap_direct=True,
            is_metal=False,
        )
        c = {field: {"$gte": 0, "$lte": 5} for field in fields}

        assert q == {
            "criteria": {
                "bandstructure.setyawan_curtarolo.magnetic_ordering": "FM",
                "bandstructure.setyawan_curtarolo.is_gap_direct": True,
                "bandstructure.setyawan_curtarolo.is_metal": False,
                **c,
            }
        }


def test_dos_data_query():
    op = DOSDataQuery()

    proj_types = [
        DOSProjectionType.total,
        DOSProjectionType.elemental,
        DOSProjectionType.orbital,
    ]

    for proj_type in proj_types:
        q = op.query(
            projection_type=proj_type,
            spin="1",
            element=Element.Si if proj_type != DOSProjectionType.total else None,
            orbital=OrbitalType.s if proj_type != DOSProjectionType.total else None,
            band_gap_min=0,
            band_gap_max=5,
            efermi_min=0,
            efermi_max=5,
            magnetic_ordering=Ordering.FM,
        )

        if proj_type == DOSProjectionType.total:
            fields = [
                "dos.total.1.band_gap",
                "dos.total.1.efermi",
            ]
        elif proj_type == DOSProjectionType.elemental:
            fields = [
                "dos.elemental.Si.s.1.band_gap",
                "dos.elemental.Si.s.1.efermi",
            ]

        elif proj_type == DOSProjectionType.orbital:
            fields = [
                "dos.orbital.s.1.band_gap",
                "dos.orbital.s.1.efermi",
            ]

        c = {field: {"$gte": 0, "$lte": 5} for field in fields}

        assert q == {"criteria": {"dos.magnetic_ordering": "FM", **c}}

        with ScratchDir("."):
            dumpfn(op, "temp.json")
            new_op = loadfn("temp.json")
            q = new_op.query(
                projection_type=proj_type,
                spin="1",
                element=Element.Si if proj_type != DOSProjectionType.total else None,
                orbital=OrbitalType.s if proj_type != DOSProjectionType.total else None,
                band_gap_min=0,
                band_gap_max=5,
                efermi_min=0,
                efermi_max=5,
                magnetic_ordering=Ordering.FM,
            )
            c = {field: {"$gte": 0, "$lte": 5} for field in fields}

            assert q == {"criteria": {"dos.magnetic_ordering": "FM", **c}}


def test_object_query():
    op = ObjectQuery()

    assert op.query(task_id="mp-149") == {"criteria": {"task_id": "mp-149"}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(task_id="mp-149") == {"criteria": {"task_id": "mp-149"}}



from monty.tempfile import ScratchDir
from monty.serialization import loadfn, dumpfn

from emmet.api.routes.molecules.query_operators import (
    MoleculeBaseQuery,
    MoleculeElementsQuery,
    MoleculeFormulaQuery,
)


def test_molecule_elements_query():
    op = MoleculeElementsQuery()

    assert op.query(elements="Si, O, P") == {
        "criteria": {"elements": {"$all": ["Si", "O", "P"]}}
    }

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(elements="Si, O, P") == {
            "criteria": {"elements": {"$all": ["Si", "O", "P"]}}
        }


def test_molecule_base_query():
    op = MoleculeBaseQuery()

    q = op.query(
        nelements_min=0,
        nelements_max=5,
        EA_min=0,
        EA_max=5,
        IE_min=0,
        IE_max=5,
        charge_min=0,
        charge_max=5,
        pointgroup="C3v",
        smiles="CN=C=O",
    )

    fields = [
        "nelements",
        "EA",
        "IE",
        "charge",
    ]

    c = {field: {"$gte": 0, "$lte": 5} for field in fields}

    assert q == {"criteria": {"pointgroup": "C3v", "smiles": "CN=C=O", **c}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        q = new_op.query(
            nelements_min=0,
            nelements_max=5,
            EA_min=0,
            EA_max=5,
            IE_min=0,
            IE_max=5,
            charge_min=0,
            charge_max=5,
            pointgroup="C3v",
            smiles="CN=C=O",
        )
        c = {field: {"$gte": 0, "$lte": 5} for field in fields}

        assert q == {"criteria": {"pointgroup": "C3v", "smiles": "CN=C=O", **c}}


def test_molecule_formula_query():
    op = MoleculeFormulaQuery()

    assert op.query(formula="C6H12O6") == {"criteria": {"formula_pretty": "H2CO"}}

    with ScratchDir("."):
        dumpfn(op, "temp.json")
        new_op = loadfn("temp.json")
        assert new_op.query(formula="C6H12O6") == {
            "criteria": {"formula_pretty": "H2CO"}
        }
